package org.apache.maven;

/* ====================================================================
 *   Copyright 2001-2004 The Apache Software Foundation.
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 * ====================================================================
 */

import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.apache.maven.project.Dependency;
import org.apache.maven.project.Project;
import org.apache.maven.repository.Artifact;
import org.apache.maven.repository.DefaultArtifactFactory;
import org.codehaus.plexus.util.StringUtils;

/**
 * @author <a href="mailto:jason@zenplex.com">Jason van Zyl</a>
 */
public class ArtifactListBuilder
{
    // --------------------------------------------------------------
    // I M P L E M E N T A T I O N
    // --------------------------------------------------------------

    /**
     * Create the list of artifacts for a project based on the stated dependencies
     * taking into account any user specified overrides.
     *
     * @param project MavenSession project.
     * @return the list of artifacts for a project
     */
    public static List build( final Project project )
    {
        List projectArtifacts = new ArrayList();
        boolean mavenJarOverride = project.getContext().getMavenJarOverride().booleanValue();

        for ( Iterator i = project.getDependencies().iterator(); i.hasNext(); )
        {
            Dependency d = (Dependency) i.next();
            String mavenJarProperty = project.getContext()
                .getMavenJarOverride( Project.standardToLegacyId( d.getId() ) );
            Artifact artifact = DefaultArtifactFactory.createArtifact( d );

            if ( mavenJarOverride && StringUtils.isNotEmpty( mavenJarProperty ) )
            {
                // The jar override option has been set and we have a property
                // for the this dependency so override the path with the user
                // specified value.
                if ( Character.isDigit( mavenJarProperty.charAt( 0 ) ) || "SNAPSHOT".equals( mavenJarProperty ) )
                {
                    // User is requesting a specific version of a dependency
                    // be used.
                    d.setVersion( mavenJarProperty );
                    artifact.setPath( project.getContext().getMavenRepoLocal() + artifact.generatePath() );
                    artifact.setOverrideType( Artifact.OVERRIDE_VERSION );
                }
                else
                {
                    // User is requesting a specific path to a dependency
                    // be used.
                    artifact.setPath( new File( mavenJarProperty ).getAbsolutePath() );
                    artifact.setOverrideType( Artifact.OVERRIDE_PATH );
                }
            }
            else
            {
                artifact.setPath( project.getContext().getMavenRepoLocal() + artifact.generatePath() );
            }

            projectArtifacts.add( artifact );
        }

        return projectArtifacts;
    }
}
