package org.apache.maven.cli;

/* ====================================================================
 *   Copyright 2001-2004 The Apache Software Foundation.
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 * ====================================================================
 */

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.PosixParser;

/**
 * Utility for managing and parsing MavenSession's command-line.
 *
 * <p>
 * <b>implementation note:</b> This is the class to modify when
 * adding, removing or modifying MavenSession's command-line options
 * and flags.
 * </p>
 *
 * <p>
 * This class uses the
 * <a href="http://jakarta.apache.org/commons/cli/">commons-cli</a>
 * library for command-line parsing.
 * </p>
 *
 * @author <a href="mailto:bob@eng.werken.com">bob mcwhirter</a>
 * @author <a href="mailto:jason@zenplex.com">Jason van Zyl</a>
 */
public class CLIManager
{
    /**
     * MavenSession't command-line option configuration.
     */
    private static Options options = null;

    /** Configure the option set. */

    static
    {
        options = new Options();

        /*
         * Although this looks broken and confusing, it is by design (commons-cli, not ours).
         * It only presents an issue if this can be called in multiple threads in the
         * same classloader simultaneously. Which it can't, as this is a static initializer.
         * The new commons-cli fixes the Builder pattern to work in a better way (OptionBuilder
         * no longer static).
         */
        OptionBuilder.withLongOpt( "nobanner" );
        OptionBuilder.withDescription( "Suppress logo banner" );
        options.addOption( OptionBuilder.create( 'b' ) );

        OptionBuilder.withLongOpt( "define" );
        OptionBuilder.hasArg();
        OptionBuilder.withDescription( "Define a system property" );
        options.addOption( OptionBuilder.create( 'D' ) );

        OptionBuilder.withLongOpt( "dir" );
        OptionBuilder.hasArg();
        OptionBuilder.withDescription( "Set effective working directory (ignored with -p or -f)" );
        options.addOption( OptionBuilder.create( 'd' ) );

        OptionBuilder.withLongOpt( "exception" );
        OptionBuilder.withDescription( "Produce exception stack traces" );
        options.addOption( OptionBuilder.create( 'e' ) );

        OptionBuilder.withLongOpt( "emacs" );
        OptionBuilder.withDescription( "Produce logging information without adornments" );
        options.addOption( OptionBuilder.create( 'E' ) );

        OptionBuilder.withLongOpt( "find" );
        OptionBuilder.hasArg();
        OptionBuilder.withDescription( "Set project file and effective working directory by finding the project file" );
        options.addOption( OptionBuilder.create( 'f' ) );

        OptionBuilder.withLongOpt( "goals" );
        OptionBuilder.withDescription( "Display available goals" );
        options.addOption( OptionBuilder.create( 'g' ) );

        OptionBuilder.withLongOpt( "usage" );
        OptionBuilder.withDescription( "Display help on using the current project" );
        options.addOption( OptionBuilder.create( 'u' ) );

        OptionBuilder.withLongOpt( "help" );
        OptionBuilder.withDescription( "Display help information" );
        options.addOption( OptionBuilder.create( 'h' ) );

        OptionBuilder.withLongOpt( "info" );
        OptionBuilder.withDescription( "Display system information" );
        options.addOption( OptionBuilder.create( 'i' ) );

        OptionBuilder.withLongOpt( "offline" );
        OptionBuilder.withDescription( "Build is happening offline" );
        options.addOption( OptionBuilder.create( 'o' ) );

        OptionBuilder.withLongOpt( "pom" );
        OptionBuilder.hasArg();
        OptionBuilder.withDescription( "Set project file" );
        options.addOption( OptionBuilder.create( 'p' ) );

        OptionBuilder.withLongOpt( "version" );
        OptionBuilder.withDescription( "Display version information" );
        options.addOption( OptionBuilder.create( 'v' ) );

        OptionBuilder.withLongOpt( "quiet" );
        OptionBuilder.withDescription( "Reduce execution output" );
        options.addOption( OptionBuilder.create( 'q' ) );

        OptionBuilder.withLongOpt( "debug" );
        OptionBuilder.withDescription( "Produce execution debug output" );
        options.addOption( OptionBuilder.create( 'X' ) );

        OptionBuilder.withLongOpt( "plugin-help" );
        OptionBuilder.withDescription( "Display help on using a given plugin" );
        OptionBuilder.hasOptionalArg();
        options.addOption( OptionBuilder.create( 'P' ) );
    }

    /**
     * Parse a string-array of command-line arguments.
     *
     * <p>
     * This will parse the arguments against the configured
     * maven command-line options, and return a <code>CommandLine</code>
     * object which may be queried for the presence of flags
     * and options and their arguments, if any.
     * </p>
     *
     * @param args The command-line arguments to parse.
     * @return The parsed <code>CommandLine</code> result.
     * @throws ParseException If an error occurs while parsing
     * the command-line options.
     * @see <a href="http://jakarta.apache.org/commons/cli/">CLI</a>
     */
    public static CommandLine parse( String[] args )
        throws ParseException
    {
        CommandLineParser parser = new PosixParser();
        return parser.parse( options, args );
    }

    /**
     * Display usage information based upon current
     * command-line option configuration.
     */
    public static void displayHelp()
    {
        HelpFormatter formatter = new HelpFormatter();

        formatter.printHelp( "maven [options] [goal [goal2 [goal3] ...]]", "\nOptions:", options, "\n" );

    }

    /**
     * Display system information. This is generally useful, maybe
     * this could be made part of CLI? But it definitely shouldn't
     * be stuck in MavenSession.
     */
    public static void displayInfo()
    {
        System.out.println( "----- Environment" );
        String[] properties = new String[] {
            "java.version",
            "file.encoding",
            "java.ext.dirs",
            "java.class.path",
            "os.name",
            "java.vendor",
            "sun.boot.class.path",
            "java.runtime.name" };
        for ( int i = 0; i < properties.length; i++ )
        {
            System.out.println( "  " + properties[i] + "=" + System.getProperty( properties[i] ) );
        }
        System.out.println( "-----" );
    }
}
