package org.apache.maven.jelly.tags.maven;

/* ====================================================================
 *   Copyright 2001-2004 The Apache Software Foundation.
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 * ====================================================================
 */

import java.io.File;
import java.io.IOException;
import java.util.StringTokenizer;

import org.apache.commons.jelly.JellyTagException;
import org.apache.commons.jelly.MissingAttributeException;
import org.apache.commons.jelly.XMLOutput;
import org.apache.maven.MavenUtils;
import org.apache.maven.jelly.tags.BaseTagSupport;

/**
 * Jelly tag to convert an absolute path into a relative path by removing the basedir.
 *
 * @author Brett Porter <a href="mailto:brett@apache.org">brett@apache.org</a>
 */
public class MakeRelativePathTag
    extends BaseTagSupport
{
    /** The base dir. */
    private File basedir;

    /** The path to convert. */
    private String path;

    /** Force a certain path separator. */
    private String separator = null;

    /** The jelly variable to store the result into. */
    private String var;

    /**
     * Set the base directory.
     * @param basedir the base directory.
     */
    public void setBasedir( File basedir )
    {
        this.basedir = basedir;
    }

    /**
     * Set the path. This has to be absolute under basedir.
     * @param path the path.
     */
    public void setPath( String path )
    {
        this.path = path;
    }

    /**
     * Set the path separator to use.
     * @param separator the separator.
     */
    public void setSeparator( String separator )
    {
        this.separator = separator;
    }

    /**
     * Set the result variable. Returns path if path is not absolute under basedir.
     * @param var the result variable name.
     */
    public void setVar( String var )
    {
        this.var = var;
    }

    /* (non-Javadoc)
     * @see org.apache.commons.jelly.Tag#doTag(org.apache.commons.jelly.XMLOutput)
     */
    public void doTag( XMLOutput output )
        throws MissingAttributeException, JellyTagException
    {
        checkAttribute( basedir, "basedir" );
        checkAttribute( path, "path" );
        checkAttribute( var, "var" );

        try
        {
            String canonicalPath = MavenUtils.makeRelativePath( basedir, path );
            if ( separator != null )
            {
                StringBuffer buf = new StringBuffer();
                StringTokenizer tok = new StringTokenizer( canonicalPath, "/\\" );
                while ( tok.hasMoreTokens() )
                {
                    buf.append( tok.nextToken() );
                    if ( tok.hasMoreTokens() )
                    {
                        buf.append( separator );
                    }
                }
                canonicalPath = buf.toString();
            }
            getContext().setVariable( var, canonicalPath );
        }
        catch ( IOException e )
        {
            throw new JellyTagException( "Unable to resolve directory", e );
        }
    }

}
