package org.apache.maven.jelly.tags.maven;

/* ====================================================================
 *   Copyright 2001-2004 The Apache Software Foundation.
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 * ====================================================================
 */

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import org.apache.commons.jelly.JellyTagException;
import org.apache.commons.jelly.MissingAttributeException;
import org.apache.commons.jelly.XMLOutput;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.maven.MavenConstants;
import org.apache.maven.MavenException;
import org.apache.maven.MavenUtils;
import org.apache.maven.project.Project;

/**
 * Reactor tag that processes a set of project descriptors taking into
 * consideration the inter-project dependencies.
 * <p/>
 * Used for building a set of projects in the correct order to satisfy any
 * dependencies among the projects being processed.
 *
 * @author <a href="mailto:bob@werken.com">bob mcwhirter</a>
 * @author <a href="mailto:jason@zenplex.com">Jason van Zyl</a>
 * @author <a href="mailto:evenisse@ifrance.com">Emmanuel Venisse</a>
 * @version $Id: ReactorTag.java 423634 2006-07-19 22:13:40Z aheritier $
 * @todo We use the basedir for the glob, but we need to set the descriptor.
 * @todo better use of inheritance from MavenTag
 */
public class ReactorTag
    extends MavenTag
{
    private static final Log log = LogFactory.getLog( ReactorTag.class );

    /**
     * The glob used to select projects within the base directory.
     */
    private String glob;

    /**
     * Projects to include.
     */
    private String includes;

    /**
     * Projects to exclude.
     */
    private String excludes;

    /**
     * Banner to display when each project is processed.
     */
    private String banner;

    /**
     * project base directory
     */
    private File basedir;

    /**
     * The name of the context variable to store the project collection in.
     */
    private String collectionVar = "reactorProjects";

    /**
     * Whether to actually execute the default or specified goals, or
     * simply collect the projects in order.
     */
    private boolean collectOnly = false;

    /**
     * Whether to sort the projects by dependencies.
     */
    private boolean sort = true;

    /**
     * Post processing flag. If this is set then we will hold on to the
     * processed projects, otherwise they will be dumped.
     */
    private boolean postProcessing = false;

    /**
     * Storage for projects that failed in the build.
     */
    private Collection failedProjects = new ArrayList();

    /**
     * List of Project objects to use instead of glob or includes.
     */
    private List projectList = null;

    // ----------------------------------------------------------------------
    // A C C E S S O R S
    // ----------------------------------------------------------------------

    /**
     * Set the project list.
     *
     * @param projectList the project list
     */
    public void setProjectList( List projectList )
    {
        this.projectList = projectList;
    }

    /**
     * Get the project list.
     *
     * @return the project list
     */
    public List getProjectList()
    {
        return projectList;
    }

    /**
     * Setter for the basedir property
     * XXX if the method it overrides is deprecated, is this also deprecated?
     *
     * @param basedir the base directory for execution of the project
     */
    public void setBasedir( File basedir )
    {
        this.basedir = basedir;
    }

    /**
     * Getter for the basedir property
     *
     * @return the base directory for execution of the project
     */
    public File getBasedir()
    {
        return this.basedir;
    }

    public void setSort( boolean sort )
    {
        this.sort = sort;
    }

    /**
     * Set the collectOnly attribute. This controls whether reactor
     * will actually execute any goals on the projects, or simply
     * collect the sorted set of projects included in this reactor
     * invocation.
     *
     * @param collectOnly if true, don't execute project goals; simply
     *                    gather the sorted project collection and store. Otherwise,
     *                    operate as before (execute either specified goals or default goal
     *                    for each project).
     */
    public void setCollectOnly( boolean collectOnly )
    {
        this.collectOnly = collectOnly;
    }

    /**
     * Set the collectionVar attribute. This controls the context
     * variable which will contain the sorted set of projects included
     * in this reactor invocation.
     *
     * @param collectionVar The context variable for the project collection
     *                      discovered here.
     */
    public void setCollectionVar( String collectionVar )
    {
        this.collectionVar = collectionVar;
    }

    /**
     * Set the postProcessing attribute.
     *
     * @param postProcessing
     */
    public void setPostProcessing( boolean postProcessing )
    {
        this.postProcessing = postProcessing;
    }

    /**
     * Get the postProcessing attribute.
     *
     * @return The
     */
    public boolean getPostProcessing()
    {
        return postProcessing;
    }

    /**
     * Set the glob used to find projects.
     *
     * @param glob The glob used to find projects.
     * @deprecated Use includes/excludes instead.
     */
    public void setGlob( String glob )
    {
        log.warn( "\nDEPRECATION WARNING: use the 'includes' attribute instead of the 'glob' attribute.\n" );
        this.glob = glob;
    }

    /**
     * Get the glob used to find projects.
     *
     * @return String The glob.
     */
    public String getGlob()
    {
        return glob;
    }

    /**
     * Set the includes used to find projects.
     *
     * @param includes The includes used to find projects.
     */
    public void setIncludes( String includes )
    {
        this.includes = includes;
    }

    /**
     * Get the includes used to find projects.
     *
     * @return String The includes.
     */
    public String getIncludes()
    {
        return includes;
    }

    /**
     * Set the excludes used to find projects.
     *
     * @param excludes The excludes used to find projects.
     */
    public void setExcludes( String excludes )
    {
        this.excludes = excludes;
    }

    /**
     * Get the excludes used to find projects.
     *
     * @return String The excludes.
     */
    public String getExcludes()
    {
        return excludes;
    }

    /**
     * Set the banner to be displayed for each project while
     * it is being processed.
     *
     * @param banner The banner to use for each processed project.
     */
    public void setBanner( String banner )
    {
        this.banner = banner;
    }

    /**
     * Get the banner to be displayed for project while it
     * is being processed.
     *
     * @return String The banner.
     */
    public String getBanner()
    {
        if ( banner == null )
        {
            banner = "Processing";
        }

        return banner;
    }

    /**
     * Execute the body of the reactor tag.
     *
     * @param output The output sink.
     * @throws JellyTagException If an error occurs while processing the tag.
     */
    public void doTag( XMLOutput output )
        throws MissingAttributeException, JellyTagException
    {
        checkAttribute( getBasedir(), "basedir" );

        if ( getGlob() == null && getIncludes() == null && projectList == null )
        {
            throw new MissingAttributeException( "glob|includes|projectList" );
        }

        log.info( "Starting the reactor..." );

        List sortedProjects = null;
        try
        {
            sortedProjects = getSortedProjects();
        }
        catch ( Exception e )
        {
            throw new JellyTagException( "Error getting projects", e );
        }

        log.info( "Our processing order:" );

        for ( Iterator i = sortedProjects.iterator(); i.hasNext(); )
        {
            Project p = (Project) i.next();
            log.info( p.getName() );
        }

        ArrayList reactorProjects = new ArrayList();

        Runtime r = Runtime.getRuntime();
        for ( Iterator i = sortedProjects.iterator(); i.hasNext(); )
        {
            // The basedir needs to be set for the project
            // We just need the descriptor.

            Project project = (Project) i.next();
            beforeProject( project );

            final long mb = 1024 * 1024;
            log.info( "+----------------------------------------" );
            log.info( "| " + getBanner() + " " + project.getName() );
            log.info( "| Memory: " + ( ( r.totalMemory() - r.freeMemory() ) / mb ) + "M/" + ( r.totalMemory() / mb )
                + "M" );
            log.info( "+----------------------------------------" );

            // We only try to attain goals if they have been set. The reactor
            // might be in use to collect project information for the purpose
            // of, say, generating a sites from a set of components.
            List goalList = null;
            if ( getGoals() != null )
            {
                goalList = MavenUtils.getGoalListFromCsv( getGoals() );
            }

            beforeLaunchGoals( project );

            try
            {
                if ( !collectOnly )
                {
                    getMavenContext().getMavenSession().attainGoals( project, goalList );
                }
            }
            catch ( Exception e )
            {
                onException( project, e );
                // TODO: there is a risk that continuing may leave the project in an inconsistent state.
                //  -- attainGoals needs to split fatal from non-fatal exceptions and we only ignore non-fatal
                if ( !isIgnoreFailures() )
                {
                    throw new JellyTagException( "Reactor subproject failure occurred", e );
                }
            }

            afterLaunchGoals( project );
            afterProject( project );

            if ( getPostProcessing() || collectOnly )
            {
                reactorProjects.add( project );
            }
        }

        getContext().setVariable( collectionVar, reactorProjects );
        Collection c = (Collection) getContext().getVariable( MavenConstants.FAILED_PROJECTS );
        if ( c == null )
        {
            c = new ArrayList( failedProjects );
        }
        else
        {
            c.addAll( failedProjects );
        }
        getContext().setVariable( MavenConstants.FAILED_PROJECTS, c );
    }

    /**
     * Get a list of projects to process, sorted by dependency processing order.
     * Maintained as a separate method so that the original projects and dependency resolver memory is freed.
     *
     * @return the project list
     */
    private List getSortedProjects()
        throws MavenException
    {
        if ( projectList != null )
        {
            log.debug( "using existing list of projects: " + projectList );
            return projectList;
        }

        String projectIncludes;
        if ( getGlob() != null )
        {
            projectIncludes = getGlob();
        }
        else
        {
            projectIncludes = getIncludes();
        }

        List projects = MavenUtils.getProjects( getBasedir(), projectIncludes, getExcludes(), getMavenContext()
            .getMavenSession().getRootContext() );

        if ( sort )
        {
            DependencyResolver dr = new DependencyResolver();
            dr.setProjects( projects );

            return dr.getSortedDependencies( false );
        }
        else
        {
            return projects;
        }
    }

    /**
     * This method is running before launching a project.
     *
     * @param project the currentProject
     */
    public void beforeProject( Project project )
    {
    }

    /**
     * This method is running before launching project goals.
     *
     * @param project the currentProject
     */
    public void beforeLaunchGoals( Project project )
    {
    }

    /**
     * This method is running after launching project goals.
     *
     * @param project the currentProject
     */
    public void afterLaunchGoals( Project project )
    {
    }

    /**
     * This method is running after launching a project.
     *
     * @param project the currentProject
     */
    public void afterProject( Project project )
    {
    }

    /**
     * This method is running when an exception occurs whatever
     * the ignoreFailures value.
     *
     * @param project the currentProject
     * @param e       the exception
     */
    public void onException( Project project, Exception e )
    {
        getContext().setVariable( MavenConstants.BUILD_FAILURE, "true" );
        failedProjects.add( project );
    }

    /**
     * This method is running when an exception occurs in
     * dependency resolution.
     *
     * @param e the exception
     */
    public void onDependencyResolutionException( Exception e )
    {
    }
}
