package org.apache.maven.jelly.tags.maven;

/* ====================================================================
 *   Copyright 2001-2004 The Apache Software Foundation.
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 * ====================================================================
 */

import java.io.File;
import java.io.IOException;
import java.util.StringTokenizer;

import org.apache.commons.jelly.JellyTagException;
import org.apache.commons.jelly.MissingAttributeException;
import org.apache.commons.jelly.Tag;
import org.apache.commons.jelly.XMLOutput;
import org.apache.maven.jelly.tags.BaseTagSupport;

/**
 * Converts an absolute path into a path relative to a root dir. For
 * example, if the root dir is "c:/apps/myproject" and the absolute
 * path is "c:/apps/myproject/path/subproject2" then the computed
 * relative path is "../..".
 *
 * @author <a href="mailto:vmassol@apache.org">Vincent Massol</a>
 * @version $Id: RootRelativePathTag.java 423634 2006-07-19 22:13:40Z aheritier $
 */
public class RootRelativePathTag
    extends BaseTagSupport
{
    /** The root dir. */
    private File rootdir;

    /** The path to convert. */
    private String path;

    /** The jelly variable to store the result into. */
    private String var;

    /**
     * Set the root directory.
     * @param rootdir the root directory
     */
    public void setRootdir( File rootdir )
    {
        this.rootdir = rootdir;
    }

    /**
     * Set the path.
     * @param path the path.
     */
    public void setPath( String path )
    {
        this.path = path;
    }

    /**
     * Set the result variable.
     * @param var the result variable name.
     */
    public void setVar( String var )
    {
        this.var = var;
    }

    /**
     * @see Tag#doTag(XMLOutput)
     */
    public void doTag( XMLOutput output )
        throws MissingAttributeException, JellyTagException
    {
        checkAttribute( rootdir, "rootdir" );
        checkAttribute( path, "path" );
        checkAttribute( var, "var" );

        String result;
        try
        {
            result = computePath();
        }
        catch ( IOException e )
        {
            throw new JellyTagException( "Unable to create relative path", e );
        }
        getContext().setVariable( var, result );
    }

    /**
     * @return the compute relative path to the root dir
     * @throws IOException on error
     */
    public String computePath()
        throws IOException
    {
        String canonicalRootdir = rootdir.getCanonicalPath();
        String canonicalPath = new File( path ).getCanonicalPath();
        if ( canonicalPath.equals( canonicalRootdir ) )
        {
            return ".";
        }
        else if ( canonicalPath.startsWith( canonicalRootdir ) )
        {
            canonicalPath = canonicalPath.substring( canonicalRootdir.length() );
        }
        else
        {
            throw new IOException( "Path [" + canonicalPath + "] is not a subset of [" + canonicalRootdir + "]" );
        }

        StringBuffer result = new StringBuffer( "." );
        StringTokenizer tokens = new StringTokenizer( canonicalPath, "/\\" );
        while ( tokens.hasMoreTokens() )
        {
            tokens.nextToken();
            result.append( "/.." );
        }
        return result.toString();
    }
}
