package org.apache.maven.util;

/* ====================================================================
 *   Copyright 2001-2004 The Apache Software Foundation.
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 * ====================================================================
 */

import java.net.URL;

/**
 * A helper bean to load the given URI from the current threads class loader or
 * the class loader that was used to load this class.
 *
 * @author <a href="mailto:jstrachan@apache.org">James Strachan</a>
 */

public class ResourceBean
{
    private ClassLoader classLoader;

    /**
     * Attempts to load the given resource from the given name.
     * The current thrad context class loader will be tried first, then the
     * class loader used to load this class.
     *
     * If the classLoader property on this bean is set then that ClassLoader is
     * used in preference to any other.
     *
     * @param name of the resource to load
     * @return URL of the resource or null if could not be found
     */
    public URL findResource( String name )
    {
        URL answer = null;
        if ( classLoader != null )
        {
            answer = classLoader.getResource( name );
        }
        // lets try the current threads class loader first
        if ( answer == null )
        {
            ClassLoader contextClassLoader = Thread.currentThread().getContextClassLoader();
            if ( contextClassLoader != null )
            {
                answer = contextClassLoader.getResource( name );
            }
            if ( answer == null )
            {
                answer = getClass().getClassLoader().getResource( name );
            }
        }
        return answer;
    }

    /**
     * Returns the classLoader.
     * @return ClassLoader
     */
    public ClassLoader getClassLoader()
    {
        return classLoader;
    }

    /**
     * Sets the classLoader.
     * @param classLoader The classLoader to set
     */
    public void setClassLoader( ClassLoader classLoader )
    {
        this.classLoader = classLoader;
    }

}
